#! /usr/bin/python

import sys
import traceback
import gobject
import logging

import time
import dbus
import dbus.mainloop.glib

TWINSTAR_IFACE	= 'com.xorcom.TwinstarCorosync'
TWINSTAR_OBJ	= '/com/xorcom/TwinstarCorosync'
TWINSTAR_DEST	= 'com.xorcom.TwinstarCorosync'
FD_IFACE        = 'org.freedesktop.DBus.Properties'

# Simplified version of:
#   https://chromium.googlesource.com/chromiumos/third_party/autotest/+/master/client/cros/mainloop.py
def ExceptionQuitMainloop(logger_name):
	def ExceptionQuitMainloop_1(func):
		def wrapper(self, *args, **kwargs):
			try:
				return func(self, *args, **kwargs)
			except Exception, e:
				logger = logging.getLogger(logger_name)
				logger.error("Communication with twinstar-corosync-bridge is broken!")
				for line in traceback.format_exc().splitlines():
					logger.error("%s", line)
				sys.exit(1)
		return wrapper
	return ExceptionQuitMainloop_1
	

# Abstract policy
class TwinstarPolicy(object):
	# Connect to TwinstarCorosync bridge
	def __init__(self):
		self.twcs = TwinstarCorosync(self)

	# Return the bridge handle
	def corosync(self):
		return self.twcs

	# All the rest must be implemented by sub-class
	def handle_new_master(self, nodeid):
		raise NotImplementedError

	def handle_joined(self, nodeid, node_count):
		raise NotImplementedError

	def handle_left(self, nodeid, node_count):
		raise NotImplementedError

	def handle_got_message(self, msg=None):
		raise NotImplementedError

	def handle_license(self, license_ok=None):
		raise NotImplementedError

class TwinstarCorosync:
	def __init__(self, twinstar_policy):
		print "Initialize"
		self.twinstar_policy = twinstar_policy
		self.bus = dbus.SystemBus()
		self.proxy = ''
		try:
			self.proxy = self.bus.get_object(TWINSTAR_DEST, TWINSTAR_OBJ)
			self.iface = dbus.Interface(self.proxy, TWINSTAR_IFACE)
			self.iface_prop = dbus.Interface(self.proxy, FD_IFACE)

			self.connect_signal("NotifyRunstepBegin", self.handle_runstep_begin)
			self.connect_signal("NotifyRunstepEnd", self.handle_runstep_end)
			self.connect_signal("NotifyRunstepsDone", self.handle_runsteps_done)
			self.connect_signal("NotifyAstribankHotplug", self.handle_astribank_hotplug)
			self.connect_signal("NotifyKeepAlive", self.handle_keepalive)
			self.connect_signal("NotifyGotMessage", self.handle_got_message)
			self.connect_signal("NotifyLicense", self.handle_license)
			self.connect_signal("NotifyNewMaster", self.handle_new_master)
			self.connect_signal("NotifyJoined", self.handle_joined)
			self.connect_signal("NotifyLeft", self.handle_left)
			# For debugging lets make a catchall
			#self.bus.add_signal_receiver(catchall_signal_handler, interface_keyword='dbus_interface', member_keyword='member')
			self.prop = {}
			self.update_property('LocalNodeId')
			self.update_property('MasterNodeId')
			self.update_property('IsMaster')
			self.update_property('ValidLicense')
			self.update_property('AstribanksCount')
		except dbus.DBusException:
			traceback.print_exc()
			raise
		print "Initialization ready (nodeid: %d)" % self.prop['LocalNodeId']

	# Private helper methods

	def connect_signal(self, name, method):
		self.proxy.connect_to_signal(name, method, dbus_interface=TWINSTAR_IFACE)

	def update_property(self, name):
		self.prop[name] = self.iface_prop.Get(TWINSTAR_IFACE, name)

	# Return interesting information

	def local_nodeid(self):
		return self.prop['LocalNodeId']

	def master_nodeid(self):
		return self.prop['MasterNodeId']

	def is_master(self):
		return self.prop['IsMaster']

	def valid_license(self):
		return self.prop['ValidLicense']

	def astribanks_count(self):
		return self.prop['AstribanksCount']

	# Hooks (call policy methods)

	def handle_new_master(self, nodeid):
		#print "new master: %d" % nodeid
		self.update_property('MasterNodeId')
		self.update_property('IsMaster')
		self.twinstar_policy.handle_new_master(nodeid)

	def handle_joined(self, nodeid, node_count):
		#print "joined: %d (node_count=%d)" % (nodeid, node_count)
		self.twinstar_policy.handle_joined(nodeid, node_count)

	def handle_left(self, nodeid, node_count):
		#print "left: %d (node_count=%d)" % (nodeid, node_count)
		self.twinstar_policy.handle_left(nodeid, node_count)

	def handle_got_message(self, msg=None):
		#print "got message: %r" % msg
		self.twinstar_policy.handle_got_message(msg)

	def handle_license(self, license_ok=None):
		#print "got message: %r" % msg
		self.update_property('ValidLicense')
		self.twinstar_policy.handle_license(license_ok)

	def handle_astribank_hotplug(self, nodeid, count, added, hardware_id, location, devpath):
		self.update_property('AstribanksCount')
		self.twinstar_policy.handle_astribank_hotplug(nodeid, count, added, hardware_id, location, devpath)

	def handle_runstep_begin(self, step_name, step_action):
		self.twinstar_policy.handle_runstep_begin(step_name, step_action)

	def handle_runstep_end(self, step_name, step_action, step_status):
		self.twinstar_policy.handle_runstep_end(step_name, step_action, step_status)

	def handle_runsteps_done(self, errors):
		self.twinstar_policy.handle_runsteps_done(errors)

	def handle_keepalive(self, nodeid):
		self.twinstar_policy.handle_keepalive(nodeid)

	def catchall_signal_handler(self, *args, **kwargs):
		print ("Caught signal (in catchall handler) "
			+ kwargs['dbus_interface'] + "." + kwargs['member'])
		for arg in args:
			print "        " + str(arg)

	def send_keep_alive(self):
		self.iface.KeepAlive()

	def send_message(self, msg):
		self.iface.SendMessage(msg)

if __name__ == '__main__':
	dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

	# Very simple policy that just print notifications and send keep-alive
	# messages every "interval"
	class DemoPolicy(TwinstarPolicy):
		def __init__(self, interval):
				super(DemoPolicy, self).__init__()
				self.twcs = self.corosync()
				self.nodeid = self.twcs.local_nodeid()
				# Give some verbose info
				me_str = ''
				if self.twcs.is_master():
					me_str = ' [me]'
				print "I: nodeid: %u" % self.nodeid
				print "I: master: %u%s" % (self.twcs.master_nodeid(), me_str)
				print "I: license: %d" % (self.twcs.valid_license())
				print "I: send keep-alive every %d msec" % interval
				if interval > 0:
					self.timer_counter = 0;
					self.timer = gobject.timeout_add(interval, self.handle_timer)

		def handle_new_master(self, nodeid):
			me_str = ''
			if self.twcs.is_master():
				me_str = ' [me]'
			print "I: new master: %d%s" % (nodeid, me_str)

		def handle_joined(self, nodeid, node_count):
			print "I: joined: %d (node_count=%d)" % (nodeid, node_count)

		def handle_left(self, nodeid, node_count):
			print "I: left: %d (node_count=%d)" % (nodeid, node_count)

		def handle_got_message(self, msg=None):
			print "I: got_message: %s" % msg

		def handle_license(self, license_ok=None):
			if license_ok:
				print "I: license: VALID"
			else:
				print "I: license: INVALID"
			print "I: valid_license = %d" % (self.twcs.valid_license())

		def handle_astribank_hotplug(self, nodeid, count, added, hardware_id, location, devpath):
			#print "DEBUG: ", nodeid, count, added_str, hardware_id, location, devpath
			if added:
				added_str = 'ADDED'
				print "I: astribank_hotplug: node %d total %d -- %s %s -- [%s] @%s\n" % (nodeid, count, added_str, devpath, hardware_id, location)
			else:
				added_str = 'REMOVED'
				print "I: astribank_hotplug: node %d total %d -- %s %s\n" % (nodeid, count, added_str, devpath)

		def handle_runstep_begin(self, step_name, step_action):
			print "I: run_step_begin: %-15s %s" % (step_name, step_action)

		def handle_runstep_end(self, step_name, step_action, step_status):
			print "I: run_step_end: %-15s %s [%d]" % (step_name, step_action, step_status)

		def handle_runsteps_done(self, errors):
			print "I: run_steps_done: Total errors = %d" % errors

		def handle_keepalive(self, nodeid):
			print "I: keepalive from: %d" % nodeid

		def handle_timer(self):
			#print "timer"
			self.twcs.send_keep_alive()
			self.timer_counter += 1
			if self.timer_counter % 3 == 0:
				self.twcs.send_message("HELLO from node %d (counter = %d)" % (self.nodeid, self.timer_counter))
			return True


	policy = DemoPolicy(3000)	# Keep-alive in msec

	loop = gobject.MainLoop()
	loop.run()


